@@ -246,6 +246,8 @@ def _estimateTableSize(args):
 	our estimate.  If it doesn't come, we return None.
 	"""
 	for fi in iterFieldInfos(args):
+		if fi.ignoreTableStats:
+			return
 		for ud in fi.userData:
 			sizeEst = getattr(ud.parent, "nrows", None)
 			if sizeEst is not None:
@@ -272,7 +274,7 @@ def _sortLargeFirst(arg1, arg2):
 		return arg1, arg2
 
 	size1, size2 = _estimateTableSize([arg1]), _estimateTableSize([arg2])
-	
+
 	if size1 is None:
 		if size2 is None:
 			# we know nothing; don't change anything to keep the user in control
@@ -311,7 +313,8 @@ class ADQLNode(utils.AutoNode):
 	
 		- the _a_<name> class attributes you need,
 		- the type (a nonterminal from the ADQL grammar) 
-		- plus bindings if the class handles more than one symbol,
+		- bindings if the class handles more than one symbol (in which case
+		  type is ignored)
 		- a class method _getInitKWs(cls, parseResult); see below.
 		- a method flatten() -> string if you define a parsed ADQLNode.
 		- a method _polish() that is called just before the constructor is
@@ -589,9 +592,9 @@ class PlainTableRef(ColumnBearingNode):
 		self.qName = flatten(self.tableName)
 
 	def flatten(self):
-		ot = flatten(self.originalTable)
-		if ot!=self.qName:
-			literal = "%s AS %s"%(ot, flatten(self.tableName))
+		origTable = flatten(self.originalTable)
+		if origTable!=self.qName:
+			literal = "%s AS %s"%(origTable, flatten(self.tableName))
 		else:
 			literal = self.qName
 
@@ -690,7 +693,7 @@ class SetGeneratingFunction(ColumnBearingNode, TransparentMixin):
 
 	def addFieldInfos(self, context):
 		# TODO: Infer types from argument types
-		fieldinfos.FieldInfos(self, context)
+		self.fieldInfos = fieldinfos.FieldInfos(self, context)
 		self.fieldInfos.addColumn(self.name,
 			fieldinfo.FieldInfo("integer", None, None, sqlName=self.name))
 	
@@ -1013,9 +1016,15 @@ class SetOperationNode(ColumnBearingNode, TransparentMixin):
 			# Skip WithQueries -- they're not part of set operations.
 			if hasattr(child, "fieldInfos") and not isinstance(child, WithQuery):
 				if fieldInfos is None:
-					fieldInfos = child.fieldInfos
+					fieldInfos = child.fieldInfos.copy(self)
 				else:
 					fieldInfos.assertIsCompatible(child.fieldInfos)
+					for (l1, mycol), (l2, theircol) in zip(fieldInfos, child.fieldInfos):
+						if mycol.ucd != theircol.ucd:
+							mycol.ucd = ""
+						if mycol.unit != theircol.unit:
+							mycol.unit = ""
+
 		return fieldInfos
 
 	def addFieldInfos(self, context):
@@ -1051,6 +1060,16 @@ class WithQuery(SetOperationNode):
 	"""
 	type = "withQuery"
 
+	def addFieldInfos(self, context):
+		# overridden because we need to discard statistics here.  We
+		# have code that looks at table sizes (look for "optimize"),
+		# and it looks at columns' parents to do that.  These stats
+		# will be totally off in CTEs.  To avoid dis-optimisations,
+		# we note it's all off here.
+		SetOperationNode.addFieldInfos(self, context)
+		for _, fi in self.fieldInfos:
+			fi.ignoreTableStats = True
+		
 	def _polish(self):
 		self.name = self.children[0]
 		for c in self.children:
@@ -1132,11 +1151,15 @@ class QuerySpecification(TransparentNode):
 		return getattr(self.children[-1], attrName)
 
 
-class ColumnReference(FieldInfoedNode):
-# normal column references will be handled by the dispatchColumnReference
-# function below, hence the binding is missing here.
+class _BaseColumnReference(FieldInfoedNode):
+	# normal column references will be handled by the dispatchColumnReference
+	# function below; there we look at the columnReference type to figure
+	# out whether something is constant, though, and so multiple
+	# non-terminals share this.
+	# That's why we're unbinding everything here and leave it to
+	# deriving classes to set up bindings.
 	type = "columnReference"
-	bindings = ["geometryValue"]
+	bindings = []
 	_a_refName = None  # if given, a TableName instance
 	_a_name = None
 
@@ -1183,10 +1206,18 @@ class ColumnReference(FieldInfoedNode):
 		return (self.type, self.name)
 
 
-class ColumnReferenceByUCD(ColumnReference):
+class ColumnReference(_BaseColumnReference):
+	# No bindings: constructed by dispatchColumnReference
+	pass
+
+class GeometryValue(_BaseColumnReference):
+	bindings = ["geometryValue"]
+
+
+class ColumnReferenceByUCD(_BaseColumnReference):
 # these are tricky: As, when parsing, we don't know where the columns
-# might come from, we have to 
-	type = "columnReferenceByUCD"
+# might come from, we have to later figure out where to get our metadata
+# from.
 	bindings = ["columnReferenceByUCD"]
 	_a_ucdWanted = None
 
@@ -1209,7 +1240,7 @@ class ColumnReferenceByUCD(ColumnReference):
 
 		for field in ancestor.fromClause.getAllFields():
 			if fnmatch.fnmatch(field.fieldInfo.ucd, self.ucdWanted):
-				self.fieldInfo = field.fieldInfo
+				self.fieldInfo = field.fieldInfo.change()
 				self.name = self.colName = field.name
 				self.refName = None
 				break
@@ -1217,6 +1248,8 @@ class ColumnReferenceByUCD(ColumnReference):
 			raise utils.NotFoundError(self.ucdWanted, "column matching ucd",
 				"from clause")
 
+		self.fieldInfo.properties["src-expression"] = flatten(self)
+		
 
 @symbolAction("columnReference")
 def dispatchColumnReference(parseResult):
@@ -1354,6 +1387,15 @@ class DerivedColumn(FieldInfoedNode):
 	def _treeRepr(self):
 		return (self.type, self.name)
 
+	def addFieldInfo(self, context):
+		FieldInfoedNode.addFieldInfo(self, context)
+		if self.fieldInfo and self.alias:
+			self.fieldInfo = self.fieldInfo.change()
+			if hasattr(self.alias, "name"): # a QName, presumably
+				self.fieldInfo.properties["src-expression"] = self.alias.name
+			else:
+				self.fieldInfo.properties["src-expression"] = self.alias
+
 
 class QualifiedStar(ADQLNode):
 	type = "qualifiedStar"
@@ -1429,12 +1471,16 @@ def _guessNumericType(literal):
 		else:
 			type = "bigint"
 	except ValueError:
-		type = "double precision"
+		if literal.lower()=="null":
+			type = None
+		else:
+			# I'm too lazy to spell out all sorts of float.
+			type = "double precision"
 	return type
 
 
 class Factor(FieldInfoedNode, TransparentMixin):
-	"""is a factor within an SQL expression.
+	"""a factor within an SQL expression.
 
 	factors may have only one (direct) child with a field info and copy
 	this.  They can have no child with a field info, in which case they're
@@ -1452,6 +1498,8 @@ class Factor(FieldInfoedNode, TransparentMixin):
 			self.fieldInfo = fieldinfo.FieldInfo(
 				_guessNumericType("".join(self.children)), "", "")
 
+		self.fieldInfo.properties["src-expression"] = flatten(self)
+		
 
 class ArrayReference(FieldInfoedNode, TransparentMixin):
 	type = "arrayReference"
@@ -1478,6 +1526,8 @@ class ArrayReference(FieldInfoedNode, TransparentMixin):
 			tainted=True # array might actually have semantics
 			)
 
+		self.fieldInfo.properties["src-expression"] = flatten(self)
+		
 
 class CombiningFINode(FieldInfoedNode):
 	def addFieldInfo(self, context):
@@ -1496,6 +1546,9 @@ class CombiningFINode(FieldInfoedNode):
 		else:
 			self.fieldInfo = self._combineFieldInfos()
 
+		if self.fieldInfo:
+			self.fieldInfo.properties["src-expression"] = flatten(self)
+		
 
 class Term(CombiningFINode, TransparentMixin):
 	type = "term"
@@ -1551,6 +1604,8 @@ class StringValueExpression(FieldInfoedNode, TransparentMixin):
 			self.fieldInfo = fieldinfo.FieldInfo(
 				"text", "", "")
 
+		self.fieldInfo.properties["src-expression"] = flatten(self)
+
 
 class GenericValueExpression(CombiningFINode, TransparentMixin):
 	"""A container for value expressions that we don't want to look at
@@ -1598,11 +1653,12 @@ class SetFunction(TransparentMixin, FieldInfoedNode):
 	type = "setFunctionSpecification"
 
 	funcDefs = {
-		'AVG': ('%s;stat.mean', None, "double precision"),
-		'MAX': ('%s;stat.max', None, None),
-		'MIN': ('%s;stat.min', None, None),
+		'AVG': ('{u};stat.mean', None, "double precision"),
+		'STDDEV': ('stat.stdev;{u}', None, "double precision"),
+		'MAX': ('stat.max;{u}', None, None),
+		'MIN': ('stat.min;{u}', None, None),
 		'SUM': (None, None, None),
-		'COUNT': ('meta.number;%s', '', "integer"),}
+		'COUNT': ('meta.number;{u}', '', "integer"),}
 
 	def addFieldInfo(self, context):
 		funcName = self.children[0].upper()
@@ -1620,7 +1676,7 @@ class SetFunction(TransparentMixin, FieldInfoedNode):
 			# ucd of a sum is the ucd of the summands?
 			ucd = fi.ucd
 		elif fi.ucd:
-			ucd = ucdPref%(fi.ucd)
+			ucd = ucdPref.format(u=fi.ucd)
 		else:
 			# no UCD given; if we're count, we're meta.number, otherwise we
 			# don't know
@@ -1640,6 +1696,8 @@ class SetFunction(TransparentMixin, FieldInfoedNode):
 		self.fieldInfo = fieldinfo.FieldInfo(
 			newType, unit=newUnit, ucd=ucd, userData=fi.userData, tainted=fi.tainted)
 
+		self.fieldInfo.properties["src-expression"] = flatten(self)
+		
 
 class NumericValueFunction(FunctionNode):
 	"""A numeric function.
@@ -1693,6 +1751,84 @@ class NumericValueFunction(FunctionNode):
 			unit, ucd, *collectUserData(infoChildren))
 		self.fieldInfo.tainted = True
 
+		self.fieldInfo.properties["src-expression"] = flatten(self)
+		
+
+class ScalarArrayFunction(FunctionNode):
+	"""All of these are functions somehow aggregating array elements.
+	
+	For those that we know we try to infer units and UCDs.
+	"""
+	type = "scalarArrayFunction"
+
+	funcDefs = {
+		# the second column was intended for unit handling but isn't used
+		# for now.  Encode unit changes there if they become necessary.
+		'ARR_MIN': ('stat.min;{u}', None, 1),
+		'ARR_MAX': ('stat.max;{u}', None, 1),
+		'ARR_AVG': ('{u};stat.mean', None, 1),
+		'ARR_STDDEV': ('stat.stdev;{u}', None, 1),
+		'ARR_SUM': ('{u};arith.sum', None, 1),
+		'ARR_DOT': (None, None, 2),
+	}
+
+	def addFieldInfo(self, context):
+		ucdchange, unitchange, nPar = self.funcDefs.get(
+			self.funName, (None, None, 1))
+
+		infoChildren = self._getInfoChildren()
+		if len(infoChildren)!=nPar:
+			raise common.Error(
+				"Grammar vs. annotation function arity mismatch!")
+
+		if nPar==1:
+			baseFI = infoChildren[0].fieldInfo
+			mat = fieldinfo.isArray(baseFI.type)
+			if not mat:
+				raise common.Error(
+					"Scalar array function called on non-array?")
+			type = mat.group(1)
+
+		elif nPar==2:
+			type = "double precision"
+			baseFI = fieldinfo.FieldInfo(type, "", "",
+				userData=infoChildren)
+		else:
+			assert False
+
+
+		ucd, unit = "", baseFI.unit
+		if ucdchange is not None and baseFI.ucd:
+			ucd = ucdchange.format(u=baseFI.ucd)
+
+		self.fieldInfo = fieldinfo.FieldInfo(type, unit, ucd,
+			*collectUserData(infoChildren))
+
+
+class ArrayMapFunction(FunctionNode):
+	"""The arr_map extension.
+
+	arg 1 is the expression; any x in it will be the filled with the array
+	elements.  arg 2 is the array to take the elements from.
+	"""
+	type = "arrayMapFunction"
+
+	def addFieldInfo(self, context):
+		self.fieldInfo = fieldinfo.FieldInfo("double precision[]", "", "")
+
+	def _polish(self):
+		# we need to pre-annotate all column references to "x" (our unbound
+		# variable) in our first argument with a literal "x" to keep
+		# our annotation engine from looking at them
+		def preAnnotate(node):
+			for child in node.iterNodeChildren():
+				if isinstance(child, ColumnReference) and child.name=="x":
+					child.fieldInfo = fieldinfo.FieldInfo(
+						"double precision", "", "")
+				preAnnotate(child)
+
+		preAnnotate(self.args[0])
+
 
 class StringValueFunction(FunctionNode):
 	type = "stringValueFunction"
@@ -1701,6 +1837,7 @@ class StringValueFunction(FunctionNode):
 		self.fieldInfo = fieldinfo.FieldInfo("text", "", "",
 			userData=collectUserData(self._getInfoChildren())[0])
 
+		self.fieldInfo.properties["src-expression"] = flatten(self)
 
 class TimestampFunction(FunctionNode):
 	type = "timestampFunction"
@@ -1829,9 +1966,45 @@ class CoalesceExpression(FieldInfoedNode, TransparentMixin):
 		if subordinates:
 			self.fieldInfo = subordinates[0].fieldInfo.change(tainted=True)
 		else:
-			self.fieldInfo = None
+			# there's probably only nulls here; make it so we produce
+			# the most robust NULL VOTable has.
+			self.fieldInfo = fieldinfo.FieldInfo('real', None, None)
 
 
+class CaseExpression(FieldInfoedNode, TransparentMixin):
+	type = "searchedCase"
+	bindings = ["searchedCase", "simpleCase"]
+
+	def addFieldInfo(self, context):
+		results, nextIsResult = [], False
+		for c in self.children:
+			if c in {"THEN", "ELSE"}:
+				nextIsResult = 1
+			else:
+				if nextIsResult:
+					results.append(c)
+					nextIsResult = False
+
+		for res in results:
+			# we try to find a case that has proper user data (i.e., a column)
+			# attached; anything else will probably be empty info.
+			if res.fieldInfo and res.fieldInfo.userData:
+				self.fieldInfo = res.fieldInfo.change(tainted=True)
+				break
+
+		else:
+			# We probably only have literals.  Get the type from the first
+			# and run.  Should we make sure that all cases have matching types?
+			for res in results:
+				if res.fieldInfo and res.fieldInfo.type:
+					self.fieldInfo = res.fieldInfo.change(tainted=True)
+					break
+
+			else:
+				# we cannot venture any guess what we'll return.  Let's avoid
+				# wild guesses for now and just fail.
+				raise common.Error("CASE statement seems to have no results.")
+
 
 ###################### Geometry and stuff that needs morphing into real SQL
 
@@ -1892,9 +2065,11 @@ class GeometryNode(CoosysMixin, FieldInfoedNode):
 			if fis and fis[0].stc:
 				thisSystem = fis[0].stc
 
+		ignoreTableStats = False
 		for index, fi in enumerate(fis):
 			childUserData.extend(fi.userData)
 			childUnits.append(fi.unit)
+			ignoreTableStats = ignoreTableStats or fi.ignoreTableStats
 			if not context.policy.match(fi.stc, thisSystem):
 				context.errors.append("When constructing %s: Argument %d has"
 					" incompatible STC"%(self.type, index+1))
@@ -1904,7 +2079,8 @@ class GeometryNode(CoosysMixin, FieldInfoedNode):
 			unit=",".join(childUnits), 
 			ucd="", 
 			userData=tuple(childUserData), 
-			stc=thisSystem)
+			stc=thisSystem,
+			ignoreTableStats=ignoreTableStats)
 		self.fieldInfo.properties["xtype"] = self.xtype
 
 
@@ -2062,7 +2238,7 @@ class Polygon(GeometryNode):
 			for item in res["points"]:
 				if item.type!="point":
 					return res
-			# all points: mutate args to let us fall through to the split coo
+			# all points: mutate args to let us fall through to the split coup
 			# case
 			arg.type = "polygonSplitCooArgs"
 			newArgs = []
@@ -2312,6 +2488,17 @@ def registerNode(node):
 	# adding a node probably changes the grammar bindings, so we
 	# need to clear the cached grammar
 	getTreeBuildingGrammar.cache_clear()
+
+	# if a module registering a node is re-loaded, we have to remove the
+	# node it previously registered, or we'll get a symbol clash.
+	# It's not totally straightforward to figure out what the previous
+	# version was; for now, let's look at type, but given the various
+	# ways nodes might use to bind, that's potentially not enough.
+	for existing in _ADDITIONAL_NODES:
+		if existing.type==node.type:
+			_ADDITIONAL_NODES.remove(existing)
+			break
+
 	_ADDITIONAL_NODES.append(node)
 
 
